#!/usr/bin/env python3

from sysrepo_api import sysrepo_api3 as api
import vppapi_if.vppapi_if as vppif
import socket



class SysrepoVRRP(api.SysrepoAPI):

    def __init__(self,
                 app_name="vrrp",
                 logger=None,
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 journal_log=True,
                 namespace="/ucpe:config/vrrp:vrrp",
                 indexes={"virtual-routers": "id"},
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            logger=logger,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            journal_log=journal_log,
            namespace=namespace,
            indexes=indexes,
            run_as_subscriber=run_as_subscriber,
            priority=priority
        )

        self.vr_ids = []

    def get_network_mode(self):
        return self.get_item("/ucpe:config/system:system/network-mode")

    def is_vpp_running(self):
        if self.get_network_mode() != 'vpp':
            self.logger.error('uCPE is not in VPP mode')
            print("VPP is not running")
            return

    def get_startup_config(self, full_config=False, full_path=True):
        return self.get_config_datastore("startup", full_config=full_config, full_path=full_path)

    def get_running_config(self, full_config=False, full_path=True):
        return self.get_config_datastore("running", full_config=full_config, full_path=full_path)

    @staticmethod
    def _get_vrrp_id(xpath):
        start = "[id='"
        end = "']"
        id = str(xpath)[xpath.find(start) + len(start):xpath.find(end)]
        return int(id)

    @staticmethod
    def is_ipv4_or_ipv6(ip_addresses):
        ip_list = ip_addresses.split(",")
        is_ipv6 = 0
        for ip in ip_list:
            s_ip = ip.strip()
            try:
                socket.inet_aton(s_ip)
            except socket.error:
                try:
                    socket.inet_pton(socket.AF_INET6, s_ip)
                    is_ipv6 = 1
                except socket.error:
                    return False
        return is_ipv6

    @staticmethod
    def get_loop_from_data(interface):
        vrrp_api = vppif.VppApiInterface('sr-vrrp')
        data_idx = vrrp_api.search_vppif_by_hostifname(interface)
        bridge_data = vrrp_api.get_interface_bd(data_idx[1])

        sw_if_index_values = []
        for sw_if_detail in bridge_data['sw_if_details']:
            sw_if_index_values.append(sw_if_detail['sw_if_index'])

        for idx in sw_if_index_values:
            intf_name = vrrp_api.get_ifname_by_ifidx(idx)
            if intf_name.startswith('loop'):
                intf = vrrp_api.get_if_by_name(intf_name)
                return intf['ifname']

    def verify_if_all_params_are_present_running(self, vr_id):
        vr_instance = self.get_running_vr(vr_id)
        for vr_id, config in vr_instance.items():
            if all(key in config for key in ['interface', 'vip', 'priority', 'peer']):
                return True

    def verify_if_all_params_are_present_startup(self, vr_id):
        vr_instance = self.get_startup_vr(vr_id)
        for vr_id, config in vr_instance.items():
            if all(key in config for key in ['interface', 'vip', 'priority', 'peer']):
                return True

    @staticmethod
    def get_sw_if_index_by_interface_name(int_name):
        vrrp_api = vppif.VppApiInterface('sr-vrrp')
        data = vrrp_api._vpp.api.sw_interface_dump(name_filter_valid=True, name_filter=int_name)
        sw_if_index = None

        for d in data:
            if d.interface_name == int_name:
                sw_if_index = d.sw_if_index
                break

        if sw_if_index:
            return sw_if_index
        else:
            return False

    def modify_vrrp_params(self, vr_id):
        vrrp_api = vppif.VppApiInterface('sr-vrrp')
        # get details for the given VR by vr_id
        data = vrrp_api.vrrp_vr_get_details(vr_id)
        # if there is data, we have some VRs that are applied to VPP
        if data:
            # check if our vr_id is in the applied config in VPP
            vr_id_found = any(d['vr_id'] == vr_id for d in data)
            if vr_id_found:
                config_indexed = self.get_running_virtual_routers()
                if config_indexed is None:
                    config_indexed = {}
                for vr, config in sorted(config_indexed.items()):
                    if vr == vr_id:
                        priority = config.get('priority')
                        vips = config.get('vip')
                        peer = config.get('peer')
                        if vips:
                            return vr_id, priority, vips, peer

    def apply_startup_config(self):
        self.is_vpp_running()
        self.logger.info("Apply STARTUP config for VRRP")

        config_indexed = self.get_startup_virtual_routers()
        if config_indexed is None:
            config_indexed = {}
        for vr_id, config in sorted(config_indexed.items()):
            # if all mandatory parameters exist, only then start the instance
            if self.verify_if_all_params_are_present_startup(vr_id):
                intf = self.get_loop_from_data(self.get_virtual_router_param(vr_id, 'interface'))
                sw_if_index = self.get_sw_if_index_by_interface_name(intf)
                priority = config.get('priority')
                vips = config.get('vip')
                peer = config.get('peer')
                enabled = config.get('enable')
                if vips and enabled:
                    vrrp_api = vppif.VppApiInterface('sr-vrrp')
                    is_ipv6 = self.is_ipv4_or_ipv6(vips)
                    vrrp_api.vrrp_vr_add_or_del(sw_if_index, vr_id, 1, priority, vips, is_ipv6)
                    vrrp_api.vrrp_vr_peers_set(sw_if_index, vr_id, peer, is_ipv6)
                    vrrp_api.vrrp_vr_start_stop('start', sw_if_index, vr_id, is_ipv6)

        return True

    def apply_running_config(self, change_set):
        self.logger.info("Apply running config")

        r_cfg = self.get_running_config(full_path=False)
        
        self.logger.info(f'Change_set: {change_set}')

        if not change_set:
            self.logger.error(" There are no changes in change set")
            return r_cfg

        vrrp_params_dict = {}

        for change in change_set:

            if not change["xpath"].startswith(self.namespace):
                continue
            if not 'operation' in change:
                self.logger.error('Malformed change set')
                continue

            vr_id = self._get_vrrp_id(change["xpath"])
            if not vr_id:
                self.logger.error("Invalid virtual router ID")
                return None

            vrrp_api = vppif.VppApiInterface('sr-vrrp')

            self.logger.info(f"CHANGE: {change}")

            if change["operation"] is "MODIFIED" or change["operation"] is "CREATED":

                if change["xpath"].find("/interface") != -1:
                    if vr_id not in vrrp_params_dict:
                        vrrp_params_dict[vr_id] = {}
                    vrrp_params_dict[vr_id]['interface'] = change['new_value']
                    try:
                        if self.verify_if_all_params_are_present_running(vr_id):
                            # get interface and convert it to 'loop'
                            intf = self.get_loop_from_data(change['new_value'])
                            # get sw_if_index from interface
                            sw_if_index = self.get_sw_if_index_by_interface_name(intf)
                            vrrp_params = self.modify_vrrp_params(vr_id)
                            if vrrp_params:
                                # return vr_id, priority, vips, peer
                                vips_conf = vrrp_params[2]
                                priority_conf = vrrp_params[1]
                                peer_conf = vrrp_params[3]
                                # verify if ipv6 or ipv4
                                is_ipv6 = self.is_ipv4_or_ipv6(vips_conf)

                                interface = change['old_value']
                                sw_if_index_old = self.get_sw_if_index_by_interface_name(self.get_loop_from_data(interface))
                                sw_if_index_new = sw_if_index
                                # stop and delete instance first to avoid vpp hang/crash
                                vrrp_api.vrrp_vr_start_stop('stop', sw_if_index_old, vr_id, is_ipv6)
                                vrrp_api.vrrp_vr_add_or_del(sw_if_index_old, vr_id, 0, priority_conf, vips_conf, is_ipv6)
                                # set instance anew
                                priority = vrrp_params_dict[vr_id].get('priority', priority_conf)
                                vips = vrrp_params_dict[vr_id].get('vip', vips_conf)
                                peer = vrrp_params_dict[vr_id].get('peer', peer_conf)
                                is_ipv6 = self.is_ipv4_or_ipv6(vips)

                                vrrp_api.vrrp_vr_add_or_del(sw_if_index_new, vr_id, 1, priority, vips, is_ipv6)
                                vrrp_api.vrrp_vr_peers_set(sw_if_index_new, vr_id, peer, is_ipv6)
                                vrrp_api.vrrp_vr_start_stop('start', sw_if_index, vr_id, is_ipv6)
                    except KeyError:
                        self.logger.info(f"Not all params are set yet in order to enable VRRP instance with ID: {vr_id}")

                elif change["xpath"].find("/priority") != -1:
                    if vr_id not in vrrp_params_dict:
                        vrrp_params_dict[vr_id] = {}
                    vrrp_params_dict[vr_id]['priority'] = change['new_value']
                    try:
                        if self.verify_if_all_params_are_present_running(vr_id):
                            intf = self.get_loop_from_data(self.get_virtual_router_param(vr_id, 'interface'))
                            # get sw_if_index from interface
                            sw_if_index = self.get_sw_if_index_by_interface_name(intf)
                            vrrp_params = self.modify_vrrp_params(vr_id)
                            if vrrp_params:
                                vips_conf = vrrp_params[2]
                                priority_old = change['old_value']
                                priority_new = change['new_value']
                                peer_conf = vrrp_params[3]
                                # verify if ipv6 or ipv4
                                is_ipv6 = self.is_ipv4_or_ipv6(vips_conf)
                                # stop and delete instance first before applying it anew to avoid vpp hang/crash
                                vrrp_api.vrrp_vr_start_stop('stop', sw_if_index, vr_id, is_ipv6)
                                vrrp_api.vrrp_vr_add_or_del(sw_if_index, vr_id, 0, priority_old, vips_conf, is_ipv6)

                                if 'interface' in vrrp_params_dict[vr_id]:
                                    intf = self.get_loop_from_data(vrrp_params_dict[vr_id]['interface'])
                                    sw_if_index = self.get_sw_if_index_by_interface_name(intf)
                                vips = vrrp_params_dict[vr_id].get('vip', vips_conf)
                                peer = vrrp_params_dict[vr_id].get('peer', peer_conf)
                                is_ipv6 = self.is_ipv4_or_ipv6(vips)

                                vrrp_api.vrrp_vr_add_or_del(sw_if_index, vr_id, 1, priority_new, vips, is_ipv6)
                                vrrp_api.vrrp_vr_peers_set(sw_if_index, vr_id, peer, is_ipv6)
                                vrrp_api.vrrp_vr_start_stop('start', sw_if_index, vr_id, is_ipv6)
                    except KeyError:
                        self.logger.error(f"Not all params are set yet in order to enable VRRP instance with ID: {vr_id}")

                elif change["xpath"].find("/vip") != -1:
                    if vr_id not in vrrp_params_dict:
                        vrrp_params_dict[vr_id] = {}
                    vrrp_params_dict[vr_id]['vip'] = change['new_value']
                    try:
                        if self.verify_if_all_params_are_present_running(vr_id):
                            # get interface and convert it to 'loop'
                            intf = self.get_loop_from_data(self.get_virtual_router_param(vr_id, 'interface'))
                            # get sw_if_index from interface
                            sw_if_index = self.get_sw_if_index_by_interface_name(intf)
                            vrrp_params = self.modify_vrrp_params(vr_id)
                            if vrrp_params:
                                vips_old = change['old_value']
                                vips_new = change['new_value']
                                priority_conf = vrrp_params[1]
                                peer_conf = vrrp_params[3]
                                # verify if ipv6 or ipv4
                                is_ipv6 = self.is_ipv4_or_ipv6(vips_old)
                                # stop and delete instance first before applying it anew to avoid vpp hang/crash
                                vrrp_api.vrrp_vr_start_stop('stop', sw_if_index, vr_id, is_ipv6)
                                vrrp_api.vrrp_vr_add_or_del(sw_if_index, vr_id, 0, priority_conf, vips_old, is_ipv6)

                                if 'interface' in vrrp_params_dict[vr_id]:
                                    intf = self.get_loop_from_data(vrrp_params_dict[vr_id]['interface'])
                                    sw_if_index = self.get_sw_if_index_by_interface_name(intf)
                                priority = vrrp_params_dict[vr_id].get('priority', priority_conf)
                                peer = vrrp_params_dict[vr_id].get('peer', peer_conf)
                                is_ipv6 = self.is_ipv4_or_ipv6(vips_new)

                                vrrp_api.vrrp_vr_add_or_del(sw_if_index, vr_id, 1, priority, vips_new, is_ipv6)
                                vrrp_api.vrrp_vr_peers_set(sw_if_index, vr_id, peer, is_ipv6)
                                vrrp_api.vrrp_vr_start_stop('start', sw_if_index, vr_id, is_ipv6)
                    except KeyError:
                        self.logger.error(f"Not all params are set yet in order to enable VRRP instance with ID: {vr_id}")

                elif change["xpath"].find("/peer") != -1:
                    if vr_id not in vrrp_params_dict:
                        vrrp_params_dict[vr_id] = {}
                    vrrp_params_dict[vr_id]['peer'] = change['new_value']
                    try:
                        if self.verify_if_all_params_are_present_running(vr_id):
                            # get interface and convert it to 'loop'
                            intf = self.get_loop_from_data(self.get_virtual_router_param(vr_id, 'interface'))
                            # get sw_if_index from interface
                            sw_if_index = self.get_sw_if_index_by_interface_name(intf)
                            data = vrrp_api.vrrp_vr_get_details(vr_id)
                            # if there is data, we have some VRs that are applied to VPP
                            if data:
                                # check if our vr_id is in the applied config in VPP
                                vr_id_found = any(d['vr_id'] == vr_id for d in data)
                                if vr_id_found:
                                    peers_ip = change['new_value']
                                    # verify if ipv6 or ipv4
                                    is_ipv6 = self.is_ipv4_or_ipv6(peers_ip)
                                    vrrp_api.vrrp_vr_start_stop('stop', sw_if_index, vr_id, is_ipv6)

                                    if 'interface' in vrrp_params_dict[vr_id]:
                                        intf = self.get_loop_from_data(vrrp_params_dict[vr_id]['interface'])
                                        sw_if_index = self.get_sw_if_index_by_interface_name(intf)

                                    vrrp_api.vrrp_vr_peers_set(sw_if_index, vr_id, peers_ip, is_ipv6)
                                    vrrp_api.vrrp_vr_start_stop('start', sw_if_index, vr_id, is_ipv6)
                    except KeyError:
                        self.logger.error(f"Not all params are set yet in order to enable VRRP instance with ID: {vr_id}")

                elif change["xpath"].find("/enable") != -1:
                    if change['new_value'] and 'true' in str(change['new_value']).lower():
                        if vr_id not in vrrp_params_dict:
                            vrrp_params_dict[vr_id] = {}
                        vrrp_params_dict[vr_id]['enable'] = change['new_value']
                        try:
                            if self.verify_if_all_params_are_present_running(vr_id):
                                intf = self.get_loop_from_data(self.get_virtual_router_param(vr_id, 'interface'))
                                sw_if_index = self.get_sw_if_index_by_interface_name(intf)
                                config_indexed = self.get_running_virtual_routers()
                                if config_indexed is None:
                                    config_indexed = {}
                                for vr, config in sorted(config_indexed.items()):
                                    if vr == vr_id:
                                        priority_conf = config.get('priority')
                                        vips_conf = config.get('vip')
                                        peer_conf = config.get('peer')
                                        if vips_conf:
                                            if 'interface' in vrrp_params_dict[vr_id]:
                                                intf = self.get_loop_from_data(vrrp_params_dict[vr_id]['interface'])
                                                sw_if_index = self.get_sw_if_index_by_interface_name(intf)

                                            priority = vrrp_params_dict[vr_id].get('priority', priority_conf)
                                            vips = vrrp_params_dict[vr_id].get('vip', vips_conf)
                                            peer = vrrp_params_dict[vr_id].get('peer', peer_conf)
                                            is_ipv6 = self.is_ipv4_or_ipv6(vips)

                                            vrrp_api.vrrp_vr_add_or_del(sw_if_index, vr_id, 1, priority, vips, is_ipv6)
                                            vrrp_api.vrrp_vr_peers_set(sw_if_index, vr_id, peer, is_ipv6)
                                            vrrp_api.vrrp_vr_start_stop('start', sw_if_index, vr_id, is_ipv6)
                        except KeyError:
                            self.logger.error(f"Not all params are set yet in order to enable VRRP instance with ID: {vr_id}")

                    elif 'false' in change['new_value']:
                        if vr_id not in vrrp_params_dict:
                            vrrp_params_dict[vr_id] = {}
                        vrrp_params_dict[vr_id]['enable'] = change['new_value']

                        if self.verify_if_all_params_are_present_running(vr_id):
                            intf = self.get_loop_from_data(self.get_virtual_router_param(vr_id, 'interface'))
                            sw_if_index = self.get_sw_if_index_by_interface_name(intf)
                            config_indexed = self.get_running_virtual_routers()
                            if config_indexed is None:
                                config_indexed = {}
                            for vr, config in sorted(config_indexed.items()):
                                if vr == vr_id:
                                    priority = config.get('priority')
                                    vips = config.get('vip')
                                    if vips:
                                        is_ipv6 = self.is_ipv4_or_ipv6(vips)
                                        vrrp_api.vrrp_vr_start_stop('stop', sw_if_index, vr_id, is_ipv6)
                                        vrrp_api.vrrp_vr_add_or_del(sw_if_index, vr_id, 0, priority, vips, is_ipv6)

                # code below is needed for when config comes from Manager for instance that does not yet exist in config
                # we then use the saved params in vrrp_params_dict because the one in running config are not yet present
                # due to the fact that config merge happens only after the data has been processed by the subscriber
                if not self.verify_if_all_params_are_present_running(vr_id):
                    for id_value, params in vrrp_params_dict.items():
                        if vr_id == id_value:
                            if params.get('enable') and 'true' in str(params.get('enable')).lower():
                                if all(key in params for key in ['interface', 'vip', 'priority', 'peer']):
                                    intf = self.get_loop_from_data(params.get('interface'))
                                    sw_if_index = self.get_sw_if_index_by_interface_name(intf)
                                    vips = params.get('vip')
                                    peer = params.get('peer')
                                    priority = int(params.get('priority'))
                                    is_ipv6 = self.is_ipv4_or_ipv6(vips)

                                    vrrp_api.vrrp_vr_add_or_del(sw_if_index, vr_id, 1, priority, vips, is_ipv6)
                                    vrrp_api.vrrp_vr_peers_set(sw_if_index, vr_id, peer, is_ipv6)
                                    vrrp_api.vrrp_vr_start_stop('start', sw_if_index, vr_id, is_ipv6)
                                else:
                                    self.logger.info(
                                        f"Not all params are set yet in order to enable VRRP instance with ID: {vr_id}")

            if change["operation"] is "DELETED":
                if self.verify_if_all_params_are_present_running(vr_id):
                    intf = self.get_loop_from_data(self.get_virtual_router_param(vr_id, 'interface'))
                    # get sw_if_index from interface
                    sw_if_index = self.get_sw_if_index_by_interface_name(intf)
                    vrrp_params = self.modify_vrrp_params(vr_id)
                    if vrrp_params:
                        vips = vrrp_params[2]
                        priority = vrrp_params[1]
                        peer = vrrp_params[3]
                        # verify if ipv6 or ipv4
                        is_ipv6 = self.is_ipv4_or_ipv6(vips)
                        # stop and delete instance
                        vrrp_api.vrrp_vr_start_stop('stop', sw_if_index, vr_id, is_ipv6)
                        vrrp_api.vrrp_vr_add_or_del(sw_if_index, vr_id, 0, priority, vips, is_ipv6)

        return r_cfg

    def apply_config(self, change_set=None, apply_startup=False):
        self.is_vpp_running()
        self.logger.info("Apply VRRP configuration")

        if apply_startup:
            self.apply_startup_config()
        else:
            self.apply_running_config(change_set)

        return True

    def set_virtual_router(self, vr_id):
        self.set_item("{}/virtual-routers[id='{}']".format(self.namespace, vr_id), vr_id)

    def get_virtual_router(self, vr_id):
        return self.get_item("{}/virtual-routers[id='{}']".format(self.namespace, vr_id))

    def unset_virtual_router(self, vr_id):
        self.delete_item("{}/virtual-routers[id='{}']".format(self.namespace, vr_id))

    def set_virtual_router_param(self, vr_id, param, value):
        self.set_item("{}/virtual-routers[id='{}']/{}".format(self.namespace, vr_id, param), value)

    def get_virtual_router_param(self, vr_id, param):
        return self.get_item("{}/virtual-routers[id='{}']/{}".format(self.namespace, vr_id, param))

    def unset_virtual_router_param(self, vr_id, param):
        self.delete_item("{}/virtual-routers[id='{}']/{}".format(self.namespace, vr_id, param))

    def get_running_vr(self, vr_id):
        config = self.get_running_config(full_path=False)
        config_indexed = {}
        if 'virtual-routers' not in config:
            return config_indexed
        for vr in config['virtual-routers']:
            if vr['id'] == vr_id:
                config_indexed[vr['id']] = vr
        return config_indexed

    def get_startup_vr(self, vr_id):
        config = self.get_startup_config(full_config=False, full_path=False)
        config_indexed = {}
        if 'virtual-routers' not in config:
            return config_indexed
        for vr in config['virtual-routers']:
            if vr['id'] == vr_id:
                config_indexed[vr['id']] = vr
        return config_indexed

    def get_startup_virtual_routers(self):
        config = self.get_startup_config(full_config=False, full_path=False)
        config_indexed = {}
        if 'virtual-routers' not in config:
            return config_indexed
        for vr in config['virtual-routers']:
            config_indexed[vr['id']] = vr
        return config_indexed

    def get_running_virtual_routers(self):
        config = self.get_running_config(full_config=False, full_path=False)
        config_indexed = {}
        if 'virtual-routers' not in config:
            return config_indexed
        for vr in config['virtual-routers']:
            config_indexed[vr['id']] = vr
        return config_indexed

    @staticmethod
    def get_vr_state(vr_id):
        vrrp_api = vppif.VppApiInterface('sr-vrrp')
        result = vrrp_api.vrrp_vr_get_details(vr_id)
        if result:
            for data in result:
                state = data['vr_state']
                if state:
                    return state
                else:
                    return None
        else:
            return None

